@@ -14,6 +14,7 @@ use crate::{
 /// or end of the file.
 const MAX_PADDING: usize = 3;
 
+// TODO: use FxHashMap here.
 pub fn opposite_positions(mps: &[MatchedPos]) -> HashMap<LineNumber, HashSet<LineNumber>> {
     let mut res: HashMap<LineNumber, HashSet<LineNumber>> = HashMap::new();
 
@@ -24,11 +25,7 @@ pub fn opposite_positions(mps: &[MatchedPos]) -> HashMap<LineNumber, HashSet<Lin
                 opposite_pos,
                 ..
             } => {
-                for (self_span, opposite_span) in zip_repeat_shorter(&self_pos.0, &opposite_pos.0) {
-                    let opposite_lines = res.entry(self_span.line).or_insert_with(HashSet::new);
-                    opposite_lines.insert(opposite_span.line);
-                }
-                for (self_span, opposite_span) in zip_repeat_shorter(&self_pos.1, &opposite_pos.1) {
+                for (self_span, opposite_span) in zip_repeat_shorter(self_pos, opposite_pos) {
                     let opposite_lines = res.entry(self_span.line).or_insert_with(HashSet::new);
                     opposite_lines.insert(opposite_span.line);
                 }
@@ -138,8 +135,13 @@ fn pad_after(ln: LineNumber, max_line: LineNumber) -> Vec<LineNumber> {
     res
 }
 
-fn flip_tuples<Tx: Copy, Ty: Copy>(items: &[(Tx, Ty)]) -> Vec<(Ty, Tx)> {
-    items.iter().map(|(x, y)| (*y, *x)).collect()
+pub fn flip_tuple<Tx: Copy, Ty: Copy>(pair: (Tx, Ty)) -> (Ty, Tx) {
+    let (x, y) = pair;
+    (y, x)
+}
+
+pub fn flip_tuples<Tx: Copy, Ty: Copy>(items: &[(Tx, Ty)]) -> Vec<(Ty, Tx)> {
+    items.iter().copied().map(flip_tuple).collect()
 }
 
 /// Before:
@@ -153,7 +155,7 @@ fn flip_tuples<Tx: Copy, Ty: Copy>(items: &[(Tx, Ty)]) -> Vec<(Ty, Tx)> {
 /// 122    91 (closest match)
 fn after_with_opposites(
     after_lines: &[LineNumber],
-    opposite_lines: HashMap<LineNumber, HashSet<LineNumber>>,
+    opposite_lines: &HashMap<LineNumber, HashSet<LineNumber>>,
     prev_max_opposite: Option<LineNumber>,
     max_opposite: LineNumber,
 ) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
@@ -197,41 +199,43 @@ fn after_with_opposites(
     res
 }
 
-pub fn calculate_context(
+pub fn calculate_before_context(
     lines: &[(Option<LineNumber>, Option<LineNumber>)],
-    lhs_mps: &[MatchedPos],
-    rhs_mps: &[MatchedPos],
-    max_lhs_src_line: LineNumber,
-    max_rhs_src_line: LineNumber,
-) -> (
-    Vec<(Option<LineNumber>, Option<LineNumber>)>,
-    Vec<(Option<LineNumber>, Option<LineNumber>)>,
-) {
-    let opposite_to_lhs = opposite_positions(lhs_mps);
-    let opposite_to_rhs = opposite_positions(rhs_mps);
-
-    let before_lines: Vec<_> = match lines.first() {
+    opposite_to_lhs: &HashMap<LineNumber, HashSet<LineNumber>>,
+    opposite_to_rhs: &HashMap<LineNumber, HashSet<LineNumber>>,
+) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
+    match lines.first() {
         Some(first_line) => match *first_line {
             (Some(lhs_line), _) => {
                 let padded_lines = pad_before(lhs_line);
-                before_with_opposites(&padded_lines, &opposite_to_lhs)
+                before_with_opposites(&padded_lines, opposite_to_lhs)
             }
             (_, Some(rhs_line)) => {
                 let padded_lines = pad_before(rhs_line);
-                flip_tuples(&before_with_opposites(&padded_lines, &opposite_to_rhs))
+                flip_tuples(&before_with_opposites(&padded_lines, opposite_to_rhs))
             }
-            (None, None) => return (vec![], vec![]),
+            (None, None) => vec![],
         },
-        None => return (vec![], vec![]),
-    };
+        None => vec![],
+    }
+}
 
-    let after_lines = match lines.last() {
+pub fn calculate_after_context(
+    lines: &[(Option<LineNumber>, Option<LineNumber>)],
+    opposite_to_lhs: &HashMap<LineNumber, HashSet<LineNumber>>,
+    opposite_to_rhs: &HashMap<LineNumber, HashSet<LineNumber>>,
+    max_lhs_src_line: LineNumber,
+    max_rhs_src_line: LineNumber,
+) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
+    match lines.last() {
         Some(first_line) => match *first_line {
             (Some(lhs_line), _) => {
                 let mut max_opposite = None;
-                for (_, rhs_line) in [&before_lines, lines].concat() {
+                // TODO: It would be simpler to do one loop over all
+                // the lines and take the last non-None on each side.
+                for (_, rhs_line) in lines {
                     if let Some(rhs_line) = rhs_line {
-                        max_opposite = Some(rhs_line);
+                        max_opposite = Some(*rhs_line);
                     }
                 }
 
@@ -245,9 +249,9 @@ pub fn calculate_context(
             }
             (_, Some(rhs_line)) => {
                 let mut max_opposite = None;
-                for (lhs_line, _) in [&before_lines, lines].concat() {
+                for (lhs_line, _) in lines {
                     if let Some(lhs_line) = lhs_line {
-                        max_opposite = Some(lhs_line);
+                        max_opposite = Some(*lhs_line);
                     }
                 }
 
@@ -259,23 +263,27 @@ pub fn calculate_context(
                     max_lhs_src_line,
                 ))
             }
-            (None, None) => return (vec![], vec![]),
+            (None, None) => vec![],
         },
-        None => return (vec![], vec![]),
-    };
-
-    (before_lines, after_lines)
+        None => vec![],
+    }
 }
 
 pub fn add_context(
     lines: &[(Option<LineNumber>, Option<LineNumber>)],
-    lhs_mps: &[MatchedPos],
-    rhs_mps: &[MatchedPos],
+    opposite_to_lhs: &HashMap<LineNumber, HashSet<LineNumber>>,
+    opposite_to_rhs: &HashMap<LineNumber, HashSet<LineNumber>>,
     max_lhs_src_line: LineNumber,
     max_rhs_src_line: LineNumber,
 ) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
-    let (before_lines, after_lines) =
-        calculate_context(lines, lhs_mps, rhs_mps, max_lhs_src_line, max_rhs_src_line);
+    let before_lines = calculate_before_context(lines, opposite_to_lhs, opposite_to_rhs);
+    let after_lines = calculate_after_context(
+        &[&before_lines, lines].concat(),
+        opposite_to_lhs,
+        opposite_to_rhs,
+        max_lhs_src_line,
+        max_rhs_src_line,
+    );
 
     before_lines
         .into_iter()
@@ -283,3 +291,25 @@ pub fn add_context(
         .chain(after_lines.into_iter())
         .collect()
 }
+
+#[cfg(test)]
+mod tests {
+    use std::iter::FromIterator;
+
+    use super::*;
+    use pretty_assertions::assert_eq;
+
+    #[test]
+    fn test_calculate_before_context() {
+        let lines = vec![(Some(1.into()), Some(1.into()))];
+
+        let mut opposite_to_lhs = HashMap::new();
+        opposite_to_lhs.insert(0.into(), HashSet::from_iter([0.into()]));
+
+        let mut opposite_to_rhs = HashMap::new();
+        opposite_to_rhs.insert(0.into(), HashSet::from_iter([0.into()]));
+
+        let res = calculate_before_context(&lines, &opposite_to_lhs, &opposite_to_rhs);
+        assert_eq!(res, vec![(Some(0.into()), Some(0.into()))]);
+    }
+}
