@@ -5,10 +5,16 @@
 /// If we exceed this, the lines are stored in separate hunks.
 const MAX_DISTANCE: usize = 4;
 
-use std::collections::{HashMap, HashSet};
+use std::{
+    collections::{HashMap, HashSet},
+    iter::FromIterator,
+};
 
 use crate::{
-    context::{add_context, calculate_context},
+    context::{
+        add_context, calculate_after_context, calculate_before_context, flip_tuple, flip_tuples,
+        opposite_positions,
+    },
     lines::LineNumber,
     syntax::{zip_pad_shorter, MatchedPos},
 };
@@ -89,7 +95,7 @@ fn fill_between(
     zip_pad_shorter(&lhs_lines, &rhs_lines)
 }
 
-pub fn extract_lines(hunk: &Hunk) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
+fn extract_lines(hunk: &Hunk) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
     let mut min_lhs = None;
     let mut min_rhs = None;
     let mut max_lhs = None;
@@ -124,8 +130,8 @@ pub fn extract_lines(hunk: &Hunk) -> Vec<(Option<LineNumber>, Option<LineNumber>
 
 pub fn merge_adjacent(
     hunks: &[Hunk],
-    lhs_mps: &[MatchedPos],
-    rhs_mps: &[MatchedPos],
+    opposite_to_lhs: &HashMap<LineNumber, HashSet<LineNumber>>,
+    opposite_to_rhs: &HashMap<LineNumber, HashSet<LineNumber>>,
     max_lhs_src_line: LineNumber,
     max_rhs_src_line: LineNumber,
 ) -> Vec<Hunk> {
@@ -140,8 +146,13 @@ pub fn merge_adjacent(
         let mut rhs_lines: HashSet<LineNumber> = HashSet::new();
 
         let lines = extract_lines(hunk);
-        let contextual_lines =
-            add_context(&lines, lhs_mps, rhs_mps, max_lhs_src_line, max_rhs_src_line);
+        let contextual_lines = add_context(
+            &lines,
+            opposite_to_lhs,
+            opposite_to_rhs,
+            max_lhs_src_line,
+            max_rhs_src_line,
+        );
         for (lhs_line, rhs_line) in contextual_lines {
             if let Some(lhs_line) = lhs_line {
                 lhs_lines.insert(lhs_line);
@@ -184,7 +195,7 @@ pub fn merge_adjacent(
     res
 }
 
-fn line_close(
+fn lines_are_close(
     max_lhs: Option<LineNumber>,
     max_rhs: Option<LineNumber>,
     line: (Option<LineNumber>, Option<LineNumber>),
@@ -266,7 +277,7 @@ fn lines_to_hunks(lines: &[(Option<LineNumber>, Option<LineNumber>)]) -> Vec<Hun
     for line in enforce_increasing(lines) {
         let (lhs_line, rhs_line) = line;
 
-        if current_hunk_lines.is_empty() || line_close(max_lhs_line, max_rhs_line, line) {
+        if current_hunk_lines.is_empty() || lines_are_close(max_lhs_line, max_rhs_line, line) {
             current_hunk_lines.push(line);
         } else {
             hunks.push(Hunk {
@@ -389,20 +400,76 @@ fn sorted_novel_positions(
     res
 }
 
-pub fn matched_pos_to_hunks(lhs_mps: &[MatchedPos], rhs_mps: &[MatchedPos]) -> Vec<Hunk> {
+fn next_opposite(
+    line: LineNumber,
+    opposites: &HashMap<LineNumber, HashSet<LineNumber>>,
+    prev_opposite: Option<LineNumber>,
+) -> Option<LineNumber> {
+    opposites.get(&line).and_then(|lines_set| {
+        let mut lines: Vec<LineNumber> = lines_set.iter().copied().collect();
+        lines.sort_unstable();
+
+        lines.into_iter().find(|ln| {
+            if let Some(prev_opposite) = prev_opposite {
+                *ln > prev_opposite
+            } else {
+                true
+            }
+        })
+    })
+}
+
+fn matched_novel_lines(
+    lhs_mps: &[MatchedPos],
+    rhs_mps: &[MatchedPos],
+) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
+    let mut highest_lhs: Option<LineNumber> = None;
+    let mut highest_rhs: Option<LineNumber> = None;
+
+    let opposite_to_lhs = opposite_positions(lhs_mps);
+    let opposite_to_rhs = opposite_positions(rhs_mps);
+
     let mut lines: Vec<(Option<LineNumber>, Option<LineNumber>)> = vec![];
     for (side, mp) in sorted_novel_positions(lhs_mps, rhs_mps) {
         let self_line = mp.pos.line;
-        let opposite_line = mp.kind.first_opposite_span().map(|span| span.line);
 
-        let line = match side {
-            Side::LHS => (Some(self_line), opposite_line),
-            Side::RHS => (opposite_line, Some(self_line)),
-        };
-        lines.push(line);
+        match side {
+            Side::LHS => {
+                let should_append = if let Some(highest_lhs) = highest_lhs {
+                    self_line > highest_lhs
+                } else {
+                    true
+                };
+                if should_append {
+                    lines.push((
+                        Some(self_line),
+                        next_opposite(self_line, &opposite_to_lhs, highest_rhs),
+                    ));
+                    highest_lhs = Some(self_line);
+                }
+            }
+            Side::RHS => {
+                let should_append = if let Some(highest_rhs) = highest_rhs {
+                    self_line > highest_rhs
+                } else {
+                    true
+                };
+                if should_append {
+                    lines.push((
+                        next_opposite(self_line, &opposite_to_rhs, highest_rhs),
+                        Some(self_line),
+                    ));
+                    highest_rhs = Some(self_line);
+                }
+            }
+        }
     }
 
-    lines_to_hunks(&lines)
+    lines
+}
+
+pub fn matched_pos_to_hunks(lhs_mps: &[MatchedPos], rhs_mps: &[MatchedPos]) -> Vec<Hunk> {
+    lines_to_hunks(&matched_novel_lines(lhs_mps, rhs_mps))
 }
 
 /// Ensure that we don't miss any intermediate values.
@@ -455,82 +522,6 @@ fn ensure_contiguous(
     res
 }
 
-/// Fill matched pairs from `start` to end.
-///
-/// 10 20
-/// 11 25
-/// 17 31
-///
-/// Pairs are monotonically increasing, but may have gaps.
-///
-fn fill_aligned(
-    start: (LineNumber, LineNumber),
-    end: (LineNumber, LineNumber),
-    matched_rhs_lines: &HashMap<LineNumber, HashSet<LineNumber>>,
-) -> Vec<(LineNumber, LineNumber)> {
-    let mut res = vec![];
-
-    let mut lhs_current = start.0;
-    let mut rhs_max_so_far = start.1;
-
-    let (lhs_max, rhs_max) = end;
-
-    while lhs_current < lhs_max {
-        lhs_current = (lhs_current.0 + 1).into();
-
-        let empty = HashSet::new();
-        let mut opposite_lines: Vec<LineNumber> = matched_rhs_lines
-            .get(&lhs_current)
-            .unwrap_or(&empty)
-            .iter()
-            .copied()
-            .filter(|ln| *ln > rhs_max_so_far && *ln < rhs_max)
-            .collect();
-        opposite_lines.sort();
-
-        if let Some(opposite_line) = opposite_lines.first() {
-            res.push((lhs_current, *opposite_line));
-            rhs_max_so_far = *opposite_line;
-        }
-    }
-
-    res
-}
-
-/// Find the first pair in this vec where both items are Some. Return
-/// that pair, plus all the items that occur afterwards.
-fn split_first_pair(
-    items: Vec<(Option<LineNumber>, Option<LineNumber>)>,
-) -> (
-    Option<(LineNumber, LineNumber)>,
-    Vec<(Option<LineNumber>, Option<LineNumber>)>,
-) {
-    for (i, (lhs_line, rhs_line)) in items.iter().copied().enumerate() {
-        if let (Some(lhs_line), Some(rhs_line)) = (lhs_line, rhs_line) {
-            let after_items: Vec<(Option<LineNumber>, Option<LineNumber>)> =
-                items[i..].iter().copied().collect();
-            return (Some((lhs_line, rhs_line)), after_items);
-        }
-    }
-
-    (None, items)
-}
-
-fn split_last_pair(
-    items: Vec<(Option<LineNumber>, Option<LineNumber>)>,
-) -> (
-    Option<(LineNumber, LineNumber)>,
-    Vec<(Option<LineNumber>, Option<LineNumber>)>,
-) {
-    let mut items = items;
-    items.reverse();
-
-    let (last_pair, mut before_items) = split_first_pair(items);
-    before_items.reverse();
-
-    (last_pair, before_items)
-}
-
 /// Before:
 ///
 /// 10 --
@@ -599,42 +590,142 @@ fn compact_gaps(
     res
 }
 
+/// Return the smallest item in `item_set` that is greater than `value`.
+///
+/// If `value` is None, just return the smallest value.
+fn first_greater<T: Copy + Ord>(item_set: &HashSet<T>, value: Option<T>) -> Option<T> {
+    let mut items = Vec::from_iter(item_set.iter().copied());
+    items.sort_unstable();
+
+    match value {
+        Some(value) => {
+            for item in items {
+                if item > value {
+                    return Some(item);
+                }
+            }
+
+            None
+        }
+        None => items.first().copied(),
+    }
+}
+
+/// Starting from `lhs_start`, fill in all the lines with known
+/// opposites up to `end`. The resulting vec will include both start
+/// and end, but the intervening values are all Some.
+fn fill_matched_lines(
+    lhs_start: LineNumber,
+    max_lhs_src_line: LineNumber,
+    end: (Option<LineNumber>, Option<LineNumber>),
+    opposite_to_lhs: &HashMap<LineNumber, HashSet<LineNumber>>,
+    opposite_to_rhs: &HashMap<LineNumber, HashSet<LineNumber>>,
+) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
+    let (lhs_end, rhs_end) = end;
+
+    let mut lhs_num = lhs_start;
+    let mut rhs_num: Option<LineNumber> = None;
+
+    let mut res: Vec<(Option<LineNumber>, Option<LineNumber>)> = vec![];
+    if !opposite_to_lhs.contains_key(&lhs_start) {
+        res.push((Some(lhs_start), None));
+    }
+
+    loop {
+        if lhs_num > max_lhs_src_line {
+            break;
+        }
+
+        if let Some(lhs_end) = lhs_end {
+            if lhs_num >= lhs_end {
+                break;
+            }
+        }
+
+        if let Some(rhs_line_set) = opposite_to_lhs.get(&lhs_num) {
+            rhs_num = first_greater(rhs_line_set, rhs_num);
+
+            if let Some(rhs_num) = rhs_num {
+                if let Some(rhs_end) = rhs_end {
+                    if rhs_num >= rhs_end {
+                        break;
+                    }
+                }
+
+                res.push((Some(lhs_num), Some(rhs_num)));
+            }
+        }
+
+        lhs_num = (lhs_num.0 + 1).into();
+    }
+
+    match end {
+        (Some(_), Some(_)) => {
+            res.push(end);
+        }
+        (Some(lhs_end), None) => {
+            if lhs_end != lhs_start && !opposite_to_lhs.contains_key(&lhs_end) {
+                res.push(end);
+            }
+        }
+        (None, Some(rhs_end)) => {
+            if !opposite_to_rhs.contains_key(&rhs_end) {
+                res.push(end);
+            }
+        }
+        _ => {}
+    }
+
+    res
+}
+
 pub fn aligned_lines_from_hunk(
     hunk: &Hunk,
-    lhs_mps: &[MatchedPos],
-    rhs_mps: &[MatchedPos],
+    opposite_to_lhs: &HashMap<LineNumber, HashSet<LineNumber>>,
+    opposite_to_rhs: &HashMap<LineNumber, HashSet<LineNumber>>,
     max_lhs_src_line: LineNumber,
     max_rhs_src_line: LineNumber,
-    matched_rhs_lines: &HashMap<LineNumber, HashSet<LineNumber>>,
 ) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
     let hunk_lines: Vec<(Option<LineNumber>, Option<LineNumber>)> = hunk.lines.clone();
 
-    let (before_context, after_context) = calculate_context(
-        &hunk_lines,
-        lhs_mps,
-        rhs_mps,
+    // TODO: this largely duplicates add_context().
+    let before_context = calculate_before_context(&hunk_lines, opposite_to_lhs, opposite_to_rhs);
+
+    let mut res: Vec<(Option<LineNumber>, Option<LineNumber>)> = vec![];
+    res.extend(&before_context);
+
+    let first_novel = hunk_lines[0];
+    let hunk_end = *hunk_lines.last().expect("Hunk lines should be non-empty");
+
+    let aligned_between = match first_novel {
+        (Some(lhs_start), _) => {
+            // TODO: align based on blank lines too.
+            fill_matched_lines(
+                lhs_start,
+                max_lhs_src_line,
+                hunk_end,
+                opposite_to_lhs,
+                opposite_to_rhs,
+            )
+        }
+        (_, Some(rhs_start)) => flip_tuples(&fill_matched_lines(
+            rhs_start,
+            max_rhs_src_line,
+            flip_tuple(hunk_end),
+            opposite_to_rhs,
+            opposite_to_lhs,
+        )),
+        (None, None) => unreachable!(),
+    };
+    res.extend(aligned_between);
+
+    let after_context = calculate_after_context(
+        &res,
+        opposite_to_lhs,
+        opposite_to_rhs,
         max_lhs_src_line,
         max_rhs_src_line,
     );
-
-    let (start_pair, before_context) = split_last_pair(before_context);
-    let (end_pair, after_context) = split_first_pair(after_context);
-
-    let mut res: Vec<(Option<LineNumber>, Option<LineNumber>)> = vec![];
-    res.extend(before_context);
-    if let (Some(start_pair), Some(end_pair)) = (start_pair, end_pair) {
-        // Fill lines between.
-        let aligned_between = fill_aligned(start_pair, end_pair, matched_rhs_lines);
-
-        // TODO: align based on blank lines too.
-
-        res.extend(aligned_between.iter().map(|(x, y)| (Some(*x), Some(*y))));
-    } else {
-        // We weren't able to find both a start pair and an end pair,
-        // so we can't fill between. Use the hunk lines as-is.
-        res.extend(hunk_lines);
-    }
-
     res.extend(after_context);
 
     compact_gaps(ensure_contiguous(&res))
@@ -647,6 +738,7 @@ mod tests {
         positions::SingleLineSpan,
         syntax::{MatchKind, TokenKind},
     };
+    use pretty_assertions::assert_eq;
 
     #[test]
     fn test_sorted_novel_positions_simple() {
@@ -663,22 +755,16 @@ mod tests {
         let matched_mp = MatchedPos {
             kind: MatchKind::Unchanged {
                 highlight: TokenKind::Delimiter,
-                self_pos: (
-                    vec![SingleLineSpan {
-                        line: 1.into(),
-                        start_col: 1,
-                        end_col: 2,
-                    }],
-                    vec![],
-                ),
-                opposite_pos: (
-                    vec![SingleLineSpan {
-                        line: 2.into(),
-                        start_col: 1,
-                        end_col: 2,
-                    }],
-                    vec![],
-                ),
+                self_pos: vec![SingleLineSpan {
+                    line: 1.into(),
+                    start_col: 1,
+                    end_col: 2,
+                }],
+                opposite_pos: vec![SingleLineSpan {
+                    line: 2.into(),
+                    start_col: 1,
+                    end_col: 2,
+                }],
             },
             pos: SingleLineSpan {
                 line: 1.into(),
@@ -692,4 +778,192 @@ mod tests {
 
         assert_eq!(res, vec![(Side::LHS, novel_mp)]);
     }
+
+    /// Simulate a simple diff:
+    ///
+    /// ```text
+    /// // Old
+    /// A
+    /// B
+    ///
+    /// // New
+    /// A
+    /// x
+    /// B
+    /// ```
+    #[test]
+    fn test_aligned_lines_from_hunk() {
+        let hunk = Hunk {
+            lines: vec![(None, Some(1.into()))],
+        };
+
+        let mut opposite_to_lhs = HashMap::new();
+        opposite_to_lhs.insert(0.into(), HashSet::from_iter([0.into()]));
+        opposite_to_lhs.insert(1.into(), HashSet::from_iter([2.into()]));
+
+        let mut opposite_to_rhs = HashMap::new();
+        opposite_to_rhs.insert(0.into(), HashSet::from_iter([0.into()]));
+        opposite_to_rhs.insert(2.into(), HashSet::from_iter([1.into()]));
+
+        let res = aligned_lines_from_hunk(
+            &hunk,
+            &opposite_to_lhs,
+            &opposite_to_rhs,
+            1.into(),
+            2.into(),
+        );
+        assert_eq!(
+            res,
+            vec![
+                (Some(0.into()), Some(0.into())),
+                (None, Some(1.into())),
+                (Some(1.into()), Some(2.into()))
+            ]
+        );
+    }
+
+    /// Simulate a simple diff:
+    ///
+    /// ```text
+    /// // Old
+    /// A
+    /// B
+    ///
+    /// // New
+    /// A
+    /// x
+    /// B
+    /// y
+    /// ```
+    #[test]
+    fn test_aligned_lines_trailing_novel() {
+        let hunk = Hunk {
+            lines: vec![(None, Some(1.into())), (None, Some(3.into()))],
+        };
+
+        let mut opposite_to_lhs = HashMap::new();
+        opposite_to_lhs.insert(0.into(), HashSet::from_iter([0.into()]));
+        opposite_to_lhs.insert(1.into(), HashSet::from_iter([2.into()]));
+
+        let mut opposite_to_rhs = HashMap::new();
+        opposite_to_rhs.insert(0.into(), HashSet::from_iter([0.into()]));
+        opposite_to_rhs.insert(2.into(), HashSet::from_iter([1.into()]));
+
+        let res = aligned_lines_from_hunk(
+            &hunk,
+            &opposite_to_lhs,
+            &opposite_to_rhs,
+            1.into(),
+            3.into(),
+        );
+        assert_eq!(
+            res,
+            vec![
+                (Some(0.into()), Some(0.into())),
+                (None, Some(1.into())),
+                (Some(1.into()), Some(2.into())),
+                (None, Some(3.into())),
+            ]
+        );
+    }
+
+    /// Regression test for a case where inserting a single line on
+    /// the RHS caused us to repeat lines.
+    #[test]
+    fn test_aligned_lines_hunk_one_line_regression() {
+        let hunk = Hunk {
+            lines: vec![(None, Some(3.into()))],
+        };
+
+        let mut opposite_to_lhs = HashMap::new();
+        opposite_to_lhs.insert(0.into(), HashSet::from_iter([1.into()]));
+        opposite_to_lhs.insert(1.into(), HashSet::from_iter([2.into()]));
+        // No match for line 3 on RHS.
+        opposite_to_lhs.insert(2.into(), HashSet::from_iter([4.into()]));
+
+        let mut opposite_to_rhs = HashMap::new();
+        opposite_to_rhs.insert(1.into(), HashSet::from_iter([0.into()]));
+        opposite_to_lhs.insert(2.into(), HashSet::from_iter([1.into()]));
+        // No match for line 2 on RHS.
+        opposite_to_rhs.insert(4.into(), HashSet::from_iter([2.into()]));
+
+        let res = aligned_lines_from_hunk(
+            &hunk,
+            &opposite_to_lhs,
+            &opposite_to_rhs,
+            2.into(),
+            4.into(),
+        );
+        assert_eq!(
+            res,
+            vec![
+                (None, Some(0.into())),
+                (Some(0.into()), Some(1.into())),
+                (None, Some(2.into())),
+                // Choosing to align RHS 3 despite it not being
+                // present in opposite_to_rhs is an arbitrary choice
+                // due to compact_gaps().
+                //
+                // The only important thing to check here is just that
+                // nodes are monotonically increasing.
+                (Some(1.into()), Some(3.into())),
+                (Some(2.into()), Some(4.into())),
+            ]
+        );
+    }
+
+    #[test]
+    fn test_matched_pos_to_hunks() {
+        let matched_pos = SingleLineSpan {
+            line: 0.into(),
+            start_col: 2,
+            end_col: 3,
+        };
+
+        let lhs_mps = [
+            MatchedPos {
+                kind: MatchKind::Novel {
+                    highlight: TokenKind::Delimiter,
+                },
+                pos: SingleLineSpan {
+                    line: 0.into(),
+                    start_col: 1,
+                    end_col: 2,
+                },
+            },
+            MatchedPos {
+                kind: MatchKind::Unchanged {
+                    highlight: TokenKind::Delimiter,
+                    self_pos: vec![matched_pos],
+                    opposite_pos: vec![matched_pos],
+                },
+                pos: matched_pos,
+            },
+        ];
+
+        let rhs_mps = [
+            MatchedPos {
+                kind: MatchKind::Novel {
+                    highlight: TokenKind::Delimiter,
+                },
+                pos: SingleLineSpan {
+                    line: 0.into(),
+                    start_col: 1,
+                    end_col: 2,
+                },
+            },
+            MatchedPos {
+                kind: MatchKind::Unchanged {
+                    highlight: TokenKind::Delimiter,
+                    self_pos: vec![matched_pos],
+                    opposite_pos: vec![matched_pos],
+                },
+                pos: matched_pos,
+            },
+        ];
+
+        let hunks = matched_pos_to_hunks(&lhs_mps, &rhs_mps);
+        assert_eq!(hunks.len(), 1);
+        assert_eq!(hunks[0].lines, vec![(Some(0.into()), Some(0.into()))]);
+    }
 }
