@@ -24,6 +24,7 @@ mod style;
 mod summary;
 mod syntax;
 mod tree_sitter_parser;
+mod unchanged;
 
 #[macro_use]
 extern crate log;
@@ -47,7 +48,9 @@ use clap::{crate_version, App, AppSettings, Arg};
 use sliders::fix_all_sliders;
 use std::{env, path::Path};
 use summary::DiffResult;
+use syntax::init_next_prev;
 use typed_arena::Arena;
+use unchanged::skip_unchanged_at_ends;
 use walkdir::WalkDir;
 
 use crate::{
@@ -55,7 +58,7 @@ use crate::{
     files::{is_probably_binary, read_or_die},
     line_parser as lp,
     lines::MaxLine,
-    syntax::{change_positions, init_info},
+    syntax::{change_positions, init_all_info},
     tree_sitter_parser as tsp,
 };
 
@@ -219,7 +222,7 @@ fn main() {
                     let ts_lang = tsp::from_language(lang);
                     let arena = Arena::new();
                     let ast = tsp::parse(&arena, &src, &ts_lang);
-                    init_info(&ast, &[]);
+                    init_all_info(&ast, &[]);
                     println!("{:#?}", ast);
                 }
                 None => {
@@ -290,7 +293,7 @@ fn diff_file_content(display_path: &str, lhs_bytes: &[u8], rhs_bytes: &[u8]) ->
         return DiffResult {
             path: display_path.into(),
             language: ts_lang.map(|l| l.name.into()),
-            binary: true,
+            binary: false,
             lhs_src: "".into(),
             rhs_src: "".into(),
             lhs_positions: vec![],
@@ -304,11 +307,19 @@ fn diff_file_content(display_path: &str, lhs_bytes: &[u8], rhs_bytes: &[u8]) ->
             let lhs = tsp::parse(&arena, &lhs_src, &ts_lang);
             let rhs = tsp::parse(&arena, &rhs_src, &ts_lang);
 
-            init_info(&lhs, &rhs);
-            mark_syntax(lhs.get(0).copied(), rhs.get(0).copied());
+            init_all_info(&lhs, &rhs);
 
-            fix_all_sliders(&lhs);
-            fix_all_sliders(&rhs);
+            let (possibly_changed_lhs, possibly_changed_rhs) = skip_unchanged_at_ends(&lhs, &rhs);
+            init_next_prev(&possibly_changed_lhs);
+            init_next_prev(&possibly_changed_rhs);
+
+            mark_syntax(
+                possibly_changed_lhs.get(0).copied(),
+                possibly_changed_rhs.get(0).copied(),
+            );
+
+            fix_all_sliders(&possibly_changed_lhs);
+            fix_all_sliders(&possibly_changed_rhs);
 
             let lhs_positions = change_positions(&lhs_src, &rhs_src, &lhs);
             let rhs_positions = change_positions(&rhs_src, &lhs_src, &rhs);
@@ -424,5 +435,6 @@ mod tests {
 
         assert_eq!(res.lhs_positions, vec![]);
         assert_eq!(res.rhs_positions, vec![]);
+        assert!(!res.binary);
     }
 }
