@@ -31,18 +31,12 @@ fn display_width() -> usize {
     term_size::dimensions().map(|(w, _)| w).unwrap_or(80)
 }
 
-/// Split `s` by newlines, but guarantees that the output is nonempty.
-/// An empty string is considered as a single line of length zero.
-fn split_lines_nonempty(s: &str) -> Vec<String> {
-    if s.is_empty() {
-        return vec!["".into()];
-    }
-
-    let mut lines = vec![];
-    for line in s.lines() {
-        lines.push(line.into());
-    }
-    lines
+/// Split `s` by newlines. Always returns a non-empty vec.
+///
+/// This differs from `str::lines`, which considers `""` to be zero
+/// lines and `"foo\n"` to be one line.
+fn split_on_newlines(s: &str) -> Vec<&str> {
+    s.split('\n').collect()
 }
 
 fn format_line_num_padded(line_num: LineNumber, column_width: usize) -> String {
@@ -87,15 +81,15 @@ fn display_line_nums(
     lhs_line_num: Option<LineNumber>,
     rhs_line_num: Option<LineNumber>,
     widths: &Widths,
-    lhs_lines_with_novel: &HashSet<LineNumber>,
-    rhs_lines_with_novel: &HashSet<LineNumber>,
+    lhs_has_novel: bool,
+    rhs_has_novel: bool,
     prev_lhs_line_num: Option<LineNumber>,
     prev_rhs_line_num: Option<LineNumber>,
 ) -> (String, String) {
     let display_lhs_line_num: String = match lhs_line_num {
         Some(line_num) => {
             let s = format_line_num_padded(line_num, widths.lhs_line_nums);
-            if lhs_lines_with_novel.contains(&line_num) {
+            if lhs_has_novel {
                 s.bright_red().to_string()
             } else {
                 s
@@ -109,7 +103,7 @@ fn display_line_nums(
     let display_rhs_line_num: String = match rhs_line_num {
         Some(line_num) => {
             let s = format_line_num_padded(line_num, widths.rhs_line_nums);
-            if rhs_lines_with_novel.contains(&line_num) {
+            if rhs_has_novel {
                 s.bright_green().to_string()
             } else {
                 s
@@ -136,12 +130,9 @@ impl Widths {
     fn new(
         terminal_width: usize,
         line_nums: &[(Option<LineNumber>, Option<LineNumber>)],
-        lhs_src: &str,
-        rhs_src: &str,
+        lhs_lines: &[&str],
+        rhs_lines: &[&str],
     ) -> Self {
-        let lhs_lines = split_lines_nonempty(lhs_src);
-        let rhs_lines = split_lines_nonempty(rhs_src);
-
         let mut lhs_max_line: LineNumber = 1.into();
         let mut rhs_max_line: LineNumber = 1.into();
         let mut lhs_max_content = 1;
@@ -150,11 +141,11 @@ impl Widths {
         for (lhs_line_num, rhs_line_num) in line_nums {
             if let Some(lhs_line_num) = lhs_line_num {
                 lhs_max_line = max(lhs_max_line, *lhs_line_num);
-                lhs_max_content = max(lhs_max_content, codepoint_len(&lhs_lines[lhs_line_num.0]));
+                lhs_max_content = max(lhs_max_content, codepoint_len(lhs_lines[lhs_line_num.0]));
             }
             if let Some(rhs_line_num) = rhs_line_num {
                 rhs_max_line = max(rhs_max_line, *rhs_line_num);
-                rhs_max_content = max(rhs_max_content, codepoint_len(&rhs_lines[rhs_line_num.0]));
+                rhs_max_content = max(rhs_max_content, codepoint_len(rhs_lines[rhs_line_num.0]));
             }
         }
 
@@ -175,6 +166,77 @@ impl Widths {
     }
 }
 
+fn lines_with_novel(
+    lhs_mps: &[MatchedPos],
+    rhs_mps: &[MatchedPos],
+) -> (HashSet<LineNumber>, HashSet<LineNumber>) {
+    let lhs_lines_with_novel: HashSet<LineNumber> = lhs_mps
+        .iter()
+        .filter(|mp| mp.kind.is_change())
+        .map(|mp| mp.pos.line)
+        .collect();
+    let rhs_lines_with_novel: HashSet<LineNumber> = rhs_mps
+        .iter()
+        .filter(|mp| mp.kind.is_change())
+        .map(|mp| mp.pos.line)
+        .collect();
+
+    (lhs_lines_with_novel, rhs_lines_with_novel)
+}
+
+/// Calculate positions of highlights on both sides. This includes
+/// both syntax highlighting and added/removed content highlighting.
+fn highlight_positions(
+    lhs_mps: &[MatchedPos],
+    rhs_mps: &[MatchedPos],
+) -> (
+    HashMap<LineNumber, Vec<(SingleLineSpan, Style)>>,
+    HashMap<LineNumber, Vec<(SingleLineSpan, Style)>>,
+) {
+    let lhs_positions = color_positions(true, lhs_mps);
+    // Preallocate the hashmap assuming the average line will have 2 items on it.
+    let mut lhs_styles: HashMap<LineNumber, Vec<(SingleLineSpan, Style)>> =
+        HashMap::with_capacity(lhs_positions.len() / 2);
+    for (span, style) in lhs_positions {
+        let styles = lhs_styles.entry(span.line).or_insert_with(Vec::new);
+        styles.push((span, style));
+    }
+
+    let rhs_positions = color_positions(false, rhs_mps);
+    let mut rhs_styles: HashMap<LineNumber, Vec<(SingleLineSpan, Style)>> =
+        HashMap::with_capacity(rhs_positions.len() / 2);
+    for (span, style) in rhs_positions {
+        let styles = rhs_styles.entry(span.line).or_insert_with(Vec::new);
+        styles.push((span, style));
+    }
+
+    (lhs_styles, rhs_styles)
+}
+
+fn highlight_as_novel(
+    line_num: Option<LineNumber>,
+    lines: &[&str],
+    opposite_line_num: Option<LineNumber>,
+    lines_with_novel: &HashSet<LineNumber>,
+) -> bool {
+    if let Some(line_num) = line_num {
+        // If this line contains any novel tokens, highlight it.
+        if lines_with_novel.contains(&line_num) {
+            return true;
+        }
+
+        let line_content = lines.get(line_num.0).map(|s| str::trim(s));
+        // If this is a blank line without a corresponding line on the
+        // other side, highlight it too. This helps highlight novel
+        // blank lines.
+        if line_content == Some("") && opposite_line_num.is_none() {
+            return true;
+        }
+    }
+
+    false
+}
+
 pub fn display_hunks(
     hunks: &[Hunk],
     display_path: &str,
@@ -184,6 +246,9 @@ pub fn display_hunks(
     lhs_mps: &[MatchedPos],
     rhs_mps: &[MatchedPos],
 ) -> String {
+    let opposite_to_lhs = opposite_positions(lhs_mps);
+    let opposite_to_rhs = opposite_positions(rhs_mps);
+
     let lhs_colored_src = apply_colors(lhs_src, true, lhs_mps);
     let rhs_colored_src = apply_colors(rhs_src, false, rhs_mps);
 
@@ -199,33 +264,14 @@ pub fn display_hunks(
         return display_single_column(display_path, lang_name, &lhs_colored_src, Color::BrightRed);
     }
 
-    let mut lhs_styles: HashMap<LineNumber, Vec<(SingleLineSpan, Style)>> = HashMap::new();
-    for (span, style) in color_positions(true, lhs_mps) {
-        let styles = lhs_styles.entry(span.line).or_insert_with(Vec::new);
-        styles.push((span, style));
-    }
+    // TODO: this is largely duplicating the `apply_colors` logic.
+    let (lhs_highlights, rhs_highlights) = highlight_positions(lhs_mps, rhs_mps);
+    let lhs_lines = split_on_newlines(lhs_src);
+    let rhs_lines = split_on_newlines(rhs_src);
+    let lhs_colored_lines = split_on_newlines(&lhs_colored_src);
+    let rhs_colored_lines = split_on_newlines(&rhs_colored_src);
 
-    let mut rhs_styles: HashMap<LineNumber, Vec<(SingleLineSpan, Style)>> = HashMap::new();
-    for (span, style) in color_positions(false, rhs_mps) {
-        let styles = rhs_styles.entry(span.line).or_insert_with(Vec::new);
-        styles.push((span, style));
-    }
-
-    let lhs_lines = split_lines_nonempty(lhs_src);
-    let rhs_lines = split_lines_nonempty(rhs_src);
-    let lhs_colored_lines = split_lines_nonempty(&lhs_colored_src);
-    let rhs_colored_lines = split_lines_nonempty(&rhs_colored_src);
-
-    let lhs_lines_with_novel: HashSet<LineNumber> = lhs_mps
-        .iter()
-        .filter(|mp| mp.kind.is_change())
-        .map(|mp| mp.pos.line)
-        .collect();
-    let rhs_lines_with_novel: HashSet<LineNumber> = rhs_mps
-        .iter()
-        .filter(|mp| mp.kind.is_change())
-        .map(|mp| mp.pos.line)
-        .collect();
+    let (lhs_lines_with_novel, rhs_lines_with_novel) = lines_with_novel(lhs_mps, rhs_mps);
 
     let mut prev_lhs_line_num = None;
     let mut prev_rhs_line_num = None;
@@ -235,27 +281,38 @@ pub fn display_hunks(
     for (i, hunk) in hunks.iter().enumerate() {
         out_lines.push(style::header(display_path, i + 1, hunks.len(), lang_name));
 
-        let matched_rhs_lines = opposite_positions(lhs_mps);
         let aligned_lines = aligned_lines_from_hunk(
             hunk,
-            lhs_mps,
-            rhs_mps,
+            &opposite_to_lhs,
+            &opposite_to_rhs,
             lhs_src.max_line(),
             rhs_src.max_line(),
-            &matched_rhs_lines,
         );
         let no_lhs_changes = hunk.lines.iter().all(|(l, _)| l.is_none());
         let no_rhs_changes = hunk.lines.iter().all(|(_, r)| r.is_none());
         let same_lines = aligned_lines.iter().all(|(l, r)| l == r);
 
-        let widths = Widths::new(display_width(), &aligned_lines, lhs_src, rhs_src);
+        let widths = Widths::new(display_width(), &aligned_lines, &lhs_lines, &rhs_lines);
         for (lhs_line_num, rhs_line_num) in aligned_lines {
+            let lhs_line_novel = highlight_as_novel(
+                lhs_line_num,
+                &lhs_lines,
+                rhs_line_num,
+                &lhs_lines_with_novel,
+            );
+            let rhs_line_novel = highlight_as_novel(
+                rhs_line_num,
+                &rhs_lines,
+                lhs_line_num,
+                &rhs_lines_with_novel,
+            );
+
             let (display_lhs_line_num, display_rhs_line_num) = display_line_nums(
                 lhs_line_num,
                 rhs_line_num,
                 &widths,
-                &lhs_lines_with_novel,
-                &rhs_lines_with_novel,
+                lhs_line_novel,
+                rhs_line_novel,
                 prev_lhs_line_num,
                 prev_rhs_line_num,
             );
@@ -302,7 +359,7 @@ pub fn display_hunks(
                     Some(lhs_line_num) => split_and_apply(
                         &lhs_lines[lhs_line_num.0],
                         widths.lhs_content,
-                        lhs_styles.get(&lhs_line_num).unwrap_or(&vec![]),
+                        lhs_highlights.get(&lhs_line_num).unwrap_or(&vec![]),
                     ),
                     None => vec![" ".repeat(widths.lhs_content)],
                 };
@@ -310,7 +367,7 @@ pub fn display_hunks(
                     Some(rhs_line_num) => split_and_apply(
                         &rhs_lines[rhs_line_num.0],
                         widths.rhs_content,
-                        rhs_styles.get(&rhs_line_num).unwrap_or(&vec![]),
+                        rhs_highlights.get(&rhs_line_num).unwrap_or(&vec![]),
                     ),
                     None => vec!["".into()],
                 };
@@ -325,7 +382,8 @@ pub fn display_hunks(
                         display_lhs_line_num.clone()
                     } else {
                         let mut s = format_missing_line_num(
-                            lhs_line_num.unwrap_or_else(|| prev_lhs_line_num.unwrap_or(10.into())),
+                            lhs_line_num
+                                .unwrap_or_else(|| prev_lhs_line_num.unwrap_or_else(|| 10.into())),
                             widths.lhs_line_nums,
                         );
                         if let Some(line_num) = lhs_line_num {
@@ -339,7 +397,8 @@ pub fn display_hunks(
                         display_rhs_line_num.clone()
                     } else {
                         let mut s = format_missing_line_num(
-                            rhs_line_num.unwrap_or_else(|| prev_rhs_line_num.unwrap_or(10.into())),
+                            rhs_line_num
+                                .unwrap_or_else(|| prev_rhs_line_num.unwrap_or_else(|| 10.into())),
                             widths.rhs_line_nums,
                         );
                         if let Some(line_num) = rhs_line_num {
@@ -372,6 +431,8 @@ pub fn display_hunks(
 
 #[cfg(test)]
 mod tests {
+    use crate::syntax::{AtomKind, MatchKind, TokenKind};
+
     use super::*;
     use pretty_assertions::assert_eq;
 
@@ -381,11 +442,91 @@ mod tests {
         let widths = Widths::new(
             80,
             &line_nums,
-            "foo\nbar\n",
-            "x\nx\nx\nx\nx\nx\nx\nx\nx\nx\nx\n",
+            &split_on_newlines("foo\nbar\n"),
+            &split_on_newlines("x\nx\nx\nx\nx\nx\nx\nx\nx\nx\nx\n"),
         );
 
         assert_eq!(widths.lhs_line_nums, 2);
         assert_eq!(widths.rhs_line_nums, 3);
     }
+
+    #[test]
+    fn test_display_width() {
+        // Basic smoke test.
+        assert!(display_width() > 10);
+    }
+
+    #[test]
+    fn test_display_single_column() {
+        // Basic smoke test.
+        let res = display_single_column("foo.py", "Python", "print(123)\n", Color::Green);
+        assert!(res.len() > 10);
+    }
+
+    #[test]
+    fn test_split_line_empty() {
+        assert_eq!(split_on_newlines(""), vec![""]);
+    }
+
+    #[test]
+    fn test_split_line_single() {
+        assert_eq!(split_on_newlines("foo"), vec!["foo"]);
+    }
+
+    #[test]
+    fn test_split_line_with_newline() {
+        assert_eq!(split_on_newlines("foo\nbar"), vec!["foo", "bar"]);
+    }
+
+    #[test]
+    fn test_split_line_with_trailing_newline() {
+        assert_eq!(split_on_newlines("foo\nbar\n"), vec!["foo", "bar", ""]);
+    }
+
+    #[test]
+    fn test_display_hunks() {
+        // Simulate diffing:
+        //
+        // Old:
+        // foo
+        //
+        // New:
+        // bar
+        let lhs_mps = [MatchedPos {
+            kind: MatchKind::Novel {
+                highlight: TokenKind::Atom(AtomKind::Normal),
+            },
+            pos: SingleLineSpan {
+                line: 0.into(),
+                start_col: 0,
+                end_col: 3,
+            },
+        }];
+
+        let rhs_mps = [MatchedPos {
+            kind: MatchKind::Novel {
+                highlight: TokenKind::Atom(AtomKind::Normal),
+            },
+            pos: SingleLineSpan {
+                line: 0.into(),
+                start_col: 0,
+                end_col: 3,
+            },
+        }];
+
+        let hunks = [Hunk {
+            lines: vec![(Some(0.into()), Some(0.into()))],
+        }];
+
+        // Simple smoke test.
+        display_hunks(
+            &hunks,
+            "foo.el",
+            "Emacs Lisp",
+            "foo",
+            "bar",
+            &lhs_mps,
+            &rhs_mps,
+        );
+    }
 }
