@@ -57,16 +57,20 @@ fn split_string(s: &str, max_len: usize) -> Vec<String> {
     res
 }
 
+fn highlight_missing_style_bug(s: &str) -> String {
+    s.on_purple().to_string()
+}
+
 pub fn split_and_apply(
     line: &str,
     max_len: usize,
     styles: &[(SingleLineSpan, Style)],
 ) -> Vec<String> {
-    if styles.is_empty() {
-        // Missing styles is a bug, so higlight in purple to make this obvious.
+    if styles.is_empty() && !line.is_empty() {
+        // Missing styles is a bug, so highlight in purple to make this obvious.
         return split_string(line, max_len)
             .into_iter()
-            .map(|part| part.purple().to_string())
+            .map(|part| highlight_missing_style_bug(&part))
             .collect();
     }
 
@@ -83,16 +87,13 @@ pub fn split_and_apply(
             }
 
             if i >= prev_length {
-                // Dim text before the next span.
+                // Unstyled text before the next span.
                 if i < span.start_col {
-                    res.push_str(
-                        &substring_by_codepoint(
-                            &part,
-                            i - prev_length,
-                            span.start_col - prev_length,
-                        )
-                        .dimmed(),
-                    );
+                    res.push_str(substring_by_codepoint(
+                        &part,
+                        i - prev_length,
+                        span.start_col - prev_length,
+                    ));
                 }
             }
 
@@ -107,10 +108,10 @@ pub fn split_and_apply(
             }
             i = span.end_col;
         }
-        // Dim text after the last span.
+        // Unstyled text after the last span.
         if i < prev_length + codepoint_len(&part) {
             let span_s = substring_by_codepoint(&part, i - prev_length, codepoint_len(&part));
-            res.push_str(&span_s.dimmed());
+            res.push_str(span_s);
         }
 
         styled_parts.push(res);
@@ -123,8 +124,8 @@ pub fn split_and_apply(
 /// Return a copy of `line` with styles applied to all the spans specified.
 /// Dim any parts of the line that have no spans.
 fn apply_line(line: &str, styles: &[(SingleLineSpan, Style)]) -> String {
-    if styles.is_empty() {
-        return line.purple().to_string();
+    if styles.is_empty() && !line.is_empty() {
+        return highlight_missing_style_bug(line);
     }
 
     let mut res = String::with_capacity(line.len());
@@ -136,9 +137,9 @@ fn apply_line(line: &str, styles: &[(SingleLineSpan, Style)]) -> String {
             break;
         }
 
-        // Dim text before the next span.
+        // Unstyled text before the next span.
         if i < span.start_col {
-            res.push_str(&substring_by_codepoint(line, i, span.start_col).dimmed());
+            res.push_str(substring_by_codepoint(line, i, span.start_col));
         }
 
         // Apply style to the substring in this span.
@@ -148,10 +149,10 @@ fn apply_line(line: &str, styles: &[(SingleLineSpan, Style)]) -> String {
         i = span.end_col;
     }
 
-    // Dim text after the last span.
+    // Unstyled text after the last span.
     if i < codepoint_len(line) {
         let span_s = substring_by_codepoint(line, i, codepoint_len(line));
-        res.push_str(&span_s.dimmed());
+        res.push_str(span_s);
     }
     res
 }
@@ -193,10 +194,18 @@ pub fn color_positions(is_lhs: bool, positions: &[MatchedPos]) -> Vec<(SingleLin
         let line_pos = pos.pos;
         let style = match pos.kind {
             MatchKind::Unchanged { highlight, .. } => Style {
-                foreground: Color::White,
+                foreground: match highlight {
+                    TokenKind::Atom(AtomKind::String) => Color::Magenta,
+                    TokenKind::Atom(AtomKind::Comment) => Color::Cyan,
+                    _ => Color::White,
+                },
                 background: None,
-                bold: highlight == TokenKind::Atom(AtomKind::Keyword),
-                dimmed: highlight == TokenKind::Atom(AtomKind::Comment),
+                bold: match highlight {
+                    TokenKind::Atom(AtomKind::Keyword) => true,
+                    TokenKind::Atom(AtomKind::Type) => true,
+                    _ => false,
+                },
+                dimmed: false,
             },
             MatchKind::Novel { highlight, .. } => Style {
                 foreground: if is_lhs {
@@ -205,7 +214,12 @@ pub fn color_positions(is_lhs: bool, positions: &[MatchedPos]) -> Vec<(SingleLin
                     Color::BrightGreen
                 },
                 background: None,
-                bold: highlight == TokenKind::Atom(AtomKind::Keyword),
+                bold: match highlight {
+                    TokenKind::Delimiter => true,
+                    TokenKind::Atom(AtomKind::Keyword) => true,
+                    TokenKind::Atom(AtomKind::Type) => true,
+                    TokenKind::Atom(_) => false,
+                },
                 dimmed: false,
             },
             MatchKind::ChangedCommentPart { .. } => Style {
@@ -231,46 +245,7 @@ pub fn color_positions(is_lhs: bool, positions: &[MatchedPos]) -> Vec<(SingleLin
 }
 
 pub fn apply_colors(s: &str, is_lhs: bool, positions: &[MatchedPos]) -> String {
-    let mut styles = vec![];
-    for pos in positions {
-        let line_pos = pos.pos;
-        let style = match pos.kind {
-            MatchKind::Unchanged { highlight, .. } => Style {
-                foreground: Color::White,
-                background: None,
-                bold: highlight == TokenKind::Atom(AtomKind::Keyword),
-                dimmed: highlight == TokenKind::Atom(AtomKind::Comment),
-            },
-            MatchKind::Novel { highlight, .. } => Style {
-                foreground: if is_lhs {
-                    Color::BrightRed
-                } else {
-                    Color::BrightGreen
-                },
-                background: None,
-                bold: highlight == TokenKind::Atom(AtomKind::Keyword),
-                dimmed: false,
-            },
-            MatchKind::ChangedCommentPart { .. } => Style {
-                foreground: if is_lhs {
-                    Color::BrightRed
-                } else {
-                    Color::BrightGreen
-                },
-                background: None,
-                bold: false,
-                dimmed: false,
-            },
-            MatchKind::UnchangedCommentPart { .. } => Style {
-                foreground: if is_lhs { Color::Red } else { Color::Green },
-                background: None,
-                bold: false,
-                dimmed: false,
-            },
-        };
-        styles.push((line_pos, style));
-    }
-
+    let styles = color_positions(is_lhs, positions);
     apply(s, &styles)
 }
 
