@@ -424,7 +424,10 @@ pub fn from_language(language: guess::Language) -> TreeSitterConfig {
 }
 
 /// Parse `src` with tree-sitter.
-pub fn parse_to_tree(src: &str, config: &TreeSitterConfig) -> (tree_sitter::Tree, HashSet<usize>) {
+pub fn parse_to_tree(
+    src: &str,
+    config: &TreeSitterConfig,
+) -> (tree_sitter::Tree, HighlightedNodeIds) {
     let mut parser = ts::Parser::new();
     parser
         .set_language(config.language)
@@ -432,31 +435,65 @@ pub fn parse_to_tree(src: &str, config: &TreeSitterConfig) -> (tree_sitter::Tree
 
     let tree = parser.parse(src, None).unwrap();
 
-    let mut node_keyword_ids = HashSet::new();
-
-    let mut keyword_ish_ids = vec![];
+    let mut keyword_ish_capture_ids = vec![];
     if let Some(idx) = config.highlight_query.capture_index_for_name("keyword") {
-        keyword_ish_ids.push(idx);
+        keyword_ish_capture_ids.push(idx);
     }
     if let Some(idx) = config.highlight_query.capture_index_for_name("operator") {
-        keyword_ish_ids.push(idx);
+        keyword_ish_capture_ids.push(idx);
     }
     if let Some(idx) = config.highlight_query.capture_index_for_name("constant") {
-        keyword_ish_ids.push(idx);
+        keyword_ish_capture_ids.push(idx);
+    }
+
+    let mut string_capture_ids = vec![];
+    if let Some(idx) = config.highlight_query.capture_index_for_name("string") {
+        string_capture_ids.push(idx);
+    }
+
+    let mut type_capture_ids = vec![];
+    if let Some(idx) = config.highlight_query.capture_index_for_name("type") {
+        type_capture_ids.push(idx);
+    }
+    if let Some(idx) = config
+        .highlight_query
+        .capture_index_for_name("type.builtin")
+    {
+        type_capture_ids.push(idx);
+    }
+    if let Some(idx) = config.highlight_query.capture_index_for_name("label") {
+        // Rust uses 'label' for lifetimes, and highglighting
+        // lifetimes consistently with types seems reasonable.
+        type_capture_ids.push(idx);
+    }
+    if let Some(idx) = config.highlight_query.capture_index_for_name("tag") {
+        type_capture_ids.push(idx);
     }
 
     let mut qc = ts::QueryCursor::new();
     let q_matches = qc.matches(&config.highlight_query, tree.root_node(), src.as_bytes());
 
+    let mut keyword_ids = HashSet::new();
+    let mut string_ids = HashSet::new();
+    let mut type_ids = HashSet::new();
     for m in q_matches {
         for c in m.captures {
-            if keyword_ish_ids.contains(&c.index) {
-                node_keyword_ids.insert(c.node.id());
+            if keyword_ish_capture_ids.contains(&c.index) {
+                keyword_ids.insert(c.node.id());
+            } else if string_capture_ids.contains(&c.index) {
+                string_ids.insert(c.node.id());
+            } else if type_capture_ids.contains(&c.index) {
+                type_ids.insert(c.node.id());
             }
         }
     }
 
-    (tree, node_keyword_ids)
+    let highlights = HighlightedNodeIds {
+        keyword_ids,
+        string_ids,
+        type_ids,
+    };
+    (tree, highlights)
 }
 
 pub fn print_tree(src: &str, tree: &tree_sitter::Tree) {
@@ -500,7 +537,7 @@ pub fn parse<'a>(
         return vec![];
     }
 
-    let (tree, keyword_ids) = parse_to_tree(src, config);
+    let (tree, highlights) = parse_to_tree(src, config);
 
     let nl_pos = NewlinePositions::from(src);
     let mut cursor = tree.walk();
@@ -509,7 +546,7 @@ pub fn parse<'a>(
     // each top level syntax item.
     cursor.goto_first_child();
 
-    all_syntaxes_from_cursor(arena, src, &nl_pos, &mut cursor, config, &keyword_ids)
+    all_syntaxes_from_cursor(arena, src, &nl_pos, &mut cursor, config, &highlights)
 }
 
 fn child_tokens<'a>(src: &'a str, cursor: &mut ts::TreeCursor) -> Vec<Option<&'a str>> {
@@ -560,6 +597,12 @@ fn find_delim_positions(
     None
 }
 
+pub struct HighlightedNodeIds {
+    keyword_ids: HashSet<usize>,
+    string_ids: HashSet<usize>,
+    type_ids: HashSet<usize>,
+}
+
 /// Convert all the tree-sitter nodes at this level to difftastic
 /// syntax nodes.
 ///
@@ -570,18 +613,13 @@ fn all_syntaxes_from_cursor<'a>(
     nl_pos: &NewlinePositions,
     cursor: &mut ts::TreeCursor,
     config: &TreeSitterConfig,
-    keyword_ids: &HashSet<usize>,
+    highlights: &HighlightedNodeIds,
 ) -> Vec<&'a Syntax<'a>> {
     let mut result: Vec<&Syntax> = vec![];
 
     loop {
         result.push(syntax_from_cursor(
-            arena,
-            src,
-            nl_pos,
-            cursor,
-            config,
-            keyword_ids,
+            arena, src, nl_pos, cursor, config, highlights,
         ));
 
         if !cursor.goto_next_sibling() {
@@ -600,7 +638,7 @@ fn syntax_from_cursor<'a>(
     nl_pos: &NewlinePositions,
     cursor: &mut ts::TreeCursor,
     config: &TreeSitterConfig,
-    keyword_ids: &HashSet<usize>,
+    highlights: &HighlightedNodeIds,
 ) -> &'a Syntax<'a> {
     let node = cursor.node();
 
@@ -618,11 +656,11 @@ fn syntax_from_cursor<'a>(
     if config.atom_nodes.contains(node.kind()) {
         // Treat nodes like string literals as atoms, regardless
         // of whether they have children.
-        atom_from_cursor(arena, src, nl_pos, cursor, keyword_ids)
+        atom_from_cursor(arena, src, nl_pos, cursor, highlights)
     } else if node.child_count() > 0 {
-        list_from_cursor(arena, src, nl_pos, cursor, config, keyword_ids)
+        list_from_cursor(arena, src, nl_pos, cursor, config, highlights)
     } else {
-        atom_from_cursor(arena, src, nl_pos, cursor, keyword_ids)
+        atom_from_cursor(arena, src, nl_pos, cursor, highlights)
     }
 }
 
@@ -634,7 +672,7 @@ fn list_from_cursor<'a>(
     nl_pos: &NewlinePositions,
     cursor: &mut ts::TreeCursor,
     config: &TreeSitterConfig,
-    keyword_ids: &HashSet<usize>,
+    highlights: &HighlightedNodeIds,
 ) -> &'a Syntax<'a> {
     let root_node = cursor.node();
 
@@ -678,36 +716,21 @@ fn list_from_cursor<'a>(
         let node = cursor.node();
         if node_i < i {
             before_delim.push(syntax_from_cursor(
-                arena,
-                src,
-                nl_pos,
-                cursor,
-                config,
-                keyword_ids,
+                arena, src, nl_pos, cursor, config, highlights,
             ));
         } else if node_i == i {
             inner_open_content = &src[node.start_byte()..node.end_byte()];
             inner_open_position = nl_pos.from_offsets(node.start_byte(), node.end_byte());
         } else if node_i < j {
             between_delim.push(syntax_from_cursor(
-                arena,
-                src,
-                nl_pos,
-                cursor,
-                config,
-                keyword_ids,
+                arena, src, nl_pos, cursor, config, highlights,
             ));
         } else if node_i == j {
             inner_close_content = &src[node.start_byte()..node.end_byte()];
             inner_close_position = nl_pos.from_offsets(node.start_byte(), node.end_byte());
         } else if node_i > j {
             after_delim.push(syntax_from_cursor(
-                arena,
-                src,
-                nl_pos,
-                cursor,
-                config,
-                keyword_ids,
+                arena, src, nl_pos, cursor, config, highlights,
             ));
         }
 
@@ -758,7 +781,7 @@ fn atom_from_cursor<'a>(
     src: &str,
     nl_pos: &NewlinePositions,
     cursor: &mut ts::TreeCursor,
-    keyword_ids: &HashSet<usize>,
+    highlights: &HighlightedNodeIds,
 ) -> &'a Syntax<'a> {
     let node = cursor.node();
     let position = nl_pos.from_offsets(node.start_byte(), node.end_byte());
@@ -775,8 +798,12 @@ fn atom_from_cursor<'a>(
 
     let highlight = if node.is_extra() {
         AtomKind::Comment
-    } else if keyword_ids.contains(&node.id()) {
+    } else if highlights.keyword_ids.contains(&node.id()) {
         AtomKind::Keyword
+    } else if highlights.string_ids.contains(&node.id()) {
+        AtomKind::String
+    } else if highlights.type_ids.contains(&node.id()) {
+        AtomKind::Type
     } else {
         AtomKind::Normal
     };
